/***  IMDISP module SPECTRA.C

        SPECTRA contains the routines for spectral analysis.

          A. Warnock
          ST Systems Corp
          Mail Code 681
          NASA/Goddard Space Flight Center
          Greenbelt, MD.

     All bugs contained herein are mine and mine alone.

***/

#define __MSC

/* * * * INCLUDE files * * * */

#include <stdio.h>
#include <stdlib.h>
#include "imdef.h"
#include "imdisp.h"
#include "dispio.h"
#include "disputil.h"
#include "textutil.h"
#include "plot.h"

/* * * * External functions * * * */
extern long screen_to_DN_y(int);

/* * * * Function declarations * * * */

void   DoSpect(int, int, int, int, long *);
void   MarkPoints(int, int, int, int, long *, int *, int *, Point *,
                       Point *);
void   SumFluxes (int, int, long *, float *);
float  Baseline (int);
void   FitBaseline (void);

/* * * * Global Variables * * * */

Spectrum   baseliner;
float      slope, intercept;
Plot       ThePlot;
FILE       *dumpfile;

void DoSpect(int start, int len, int left, int right, long * buf)
/***  DoSpect has the user mark the baseline of a plotted spectrum,
      stores away the associated values, then asks the user to mark off
      a feature and sums the baseline-subtracted fluxes.

      Original version - 8/92 AW3

   Parameter    Type    In/out   Description
     start      int      in      Array index of 1st point plotted
      len       int      in      Number of array points plotted
     left       int      in      Pixel number of left side of plot
     right      int      in      Pixel number of right side of plot
      buf     long ptr   in      Array of y-values
***/
{
    int    index_low, index_high, i;
    char   dispstr[128];
    float  flux;
    Point  BaseLow, BaseHigh, LineLow, LineHigh;
    char   horiz_label[10];
    char   count_label[10];

    strcpy(horiz_label, "Pixel ");
    strcpy(count_label, "Counts ");

    dumpfile = fopen("SPECTRUM.LOG", "a");
    fprintf(dumpfile,"\n--------------------------------------\n\n");

/*  Measure the baseline for subtraction first */
    StatusLine(1,"Mark 2 points on baseline: Mouse, arrow keys move cursor");
    MarkPoints( start, len, left, right, buf, &index_low, &index_high,
                &BaseLow, &BaseHigh);
    DrawLine( BaseLow.y, BaseLow.x, BaseHigh.y, BaseHigh.x, numDN-1);
    baseliner.w_low = (long) index_low;
    baseliner.f_low = screen_to_DN_y(BaseLow.y);
/*
    baseliner.f_low = buf[index_low];
*/
    sprintf (dispstr, "%s : %8ld   %s : %8ld", horiz_label,
             baseliner.w_low, count_label, baseliner.f_low);
    StatusLine(0,dispstr);

    baseliner.w_high = (long) index_high;
    baseliner.f_high = screen_to_DN_y(BaseHigh.y);
/*
    baseliner.f_high = buf[index_high];
*/
    sprintf (dispstr, "%s : %8ld   %s : %8ld", horiz_label,
             baseliner.w_high, count_label, baseliner.f_high);
    StatusLine(3,dispstr);

    if (dumpfile != NULL)
    {
       fprintf(dumpfile,
               "Baseline: low  - line %4d, sample %4d, DN[%s %4ld] = %8ld %s\n",
               BaseLow.y, BaseLow.x, horiz_label, baseliner.w_low,
               baseliner.f_low, count_label);

       fprintf(dumpfile,
               "Baseline: high - line %4d, sample %4d, DN[%s %4ld] = %8ld %s\n",
                BaseHigh.y, BaseHigh.x, horiz_label, baseliner.w_high,
                baseliner.f_high, count_label);
    }

    FitBaseline();

/*  Now, mark off the feature to measure */
    StatusLine(1,"Mark endpoints of feature: Mouse, arrow keys move cursor");
    MarkPoints( start, len, left, right, buf, &index_low, &index_high,
                &LineLow, &LineHigh);
    sprintf( dispstr, "%s : %8d   %s : %8d", horiz_label, index_low,
             count_label, buf[index_low]);
    StatusLine(0, dispstr);
    sprintf( dispstr, "%s : %8d   %s : %8d", horiz_label, index_high,
             count_label, buf[index_high]);
    StatusLine(3, dispstr);

    if (dumpfile != NULL)
    {
       fprintf(dumpfile,
               " Feature: low  - line %4d, sample %4d, DN[%s %4d] = %8ld %s\n",
               LineLow.y, LineLow.x, horiz_label, index_low, buf[index_low],
               count_label);

       fprintf(dumpfile,
               " Feature: high - line %4d, sample %4d, DN[%s %4d] = %8ld %s\n",
               LineHigh.y, LineHigh.x, horiz_label, index_high, buf[index_high],
               count_label);
    }

    SumFluxes( index_low, index_high, buf, &flux);
    sprintf( dispstr, "Total Net %s : %12.4g", count_label, flux);
    StatusLine(1, dispstr);
    fprintf(dumpfile, "\n\n%s\n", dispstr);

    if (dumpfile != NULL) fclose(dumpfile);
}

void MarkPoints(int first, int len, int left, int right, long *
                    buf, int *start, int *end, struct Point * Low,
                    struct Point * High)
/***  MarkPoints handles allowing the user to mark the baseline of a
      plotted spectrum and stores away the associated values

      Original version - 8/92 AW3

   Parameter    Type    In/out   Description
     first      int      in      Array index of 1st point plotted
      len       int      in      Number of array points plotted
     left       int      in      Pixel number of left side of plot
     right      int      in      Pixel number of right side of plot
      buf     long ptr   in      Array of y-values
     start    int ptr    out     y-array index of 1st point marked
      end     int ptr    out     y-array index of 2nd point marked
      Low   struct ptr   out     screen coords of 1st point marked
     High   struct ptr   out     screen coords of 2nd point marked
***/
{
    int    line, samp;
    long   temp, array_end, plot_width;

    array_end = first + len - 1;
    plot_width = (long) (right - left);

/*  Put up the prompt, and get the screen coords of the first point */
    MoveCursor(&line,&samp);
    DrawText("*", line+4, samp-4, 8, 0, numDN-1);
    temp = (long) (samp - left) * len * 100L;
    temp = (temp / plot_width);
    temp = first + temp / 100L;

    *start = max(temp, first);
    Low->x = samp;
    Low->y = line;

/*  Then get the screen coords of the second point */
    MoveCursor(&line,&samp);
    DrawText("*", line+4, samp-4, 8, 0, numDN-1);
    temp = (long) (samp - left) * len * 100L;
    temp = (temp / plot_width);
    temp = first + temp / 100L;

    *end = min(temp, array_end);
    High->x = samp;
    High->y = line;

    return;
}

float Baseline( int index)
/***  Baseline calculates the (linear) fitted baseline value from
      the array index.

      Original version - 8/92 AW3

   Parameter    Type    In/out   Description
     index      int      in      Y array index
***/
{
   return( slope * (float) (index) + intercept);
}

void SumFluxes( int start, int end, long * buf, float *sum)
/***  SumFluxes calculates the sum of the flux values between two points

      Original version - 8/92 AW3

   Parameter    Type    In/out   Description
     start      int      in      Array index of starting point
      end       int      in      Array index of ending point
      buf     long ptr   in      Array of y-values
      sum    float ptr   out     Sum of fluxes
***/
{
    int i;
    float basevalue;

    *sum = 0.0;
    if (dumpfile != NULL)
        fprintf(dumpfile, "\n\nIndex      DN    Baseline (fitted)    Net          Accum\n");

    for (i=start; i<=end; i++)
    {
       basevalue = ((float) buf[i]) - Baseline(i);
       *sum = *sum + basevalue;

       if (dumpfile != NULL)
           fprintf(dumpfile, "%5d  %8ld  %12.4f  %12.4g  %12.4g\n",
                   i, buf[i], Baseline(i), basevalue, *sum);

    }
    return;
}

void FitBaseline(void)
/***  FitBaseline calculates the (linear) coefficients to relate
      the array index to the flux array.  The slope and intercept are
      global (float) variables.

      Original version - 8/92 AW3

***/
{
    long y1, y2, start, end;

    y1 = baseliner.f_low;
    y2 = baseliner.f_high;
    start = baseliner.w_low;
    end = baseliner.w_high;

    slope = ( (float) (y2 - y1)) / ( (float) (end - start));
    intercept = ( (float) y2) - slope * ( (float) end);
}
